import os
import re
from dataclasses import dataclass
from pathlib import Path
from typing import List

from nvflare.lighter.ctx import ProvisionContext
from nvflare.lighter.provision import prepare_project, prepare_builders
from nvflare.lighter.provisioner import Provisioner
from nvflare.lighter.utils import load_yaml
from nvflare.lighter.entity import Participant

from refedez.common.machines import Machine
from refedez.build.nix import Closure
from refedez.common.project import ReFedEzProject


@dataclass(frozen=True)
class FlareProvision:
    _context: ProvisionContext

    @property
    def participants(self) -> List[str]:
        participants: List[Participant] = (
            self._context.get_project().get_all_participants()
        )
        return [participant.name for participant in participants]

    def patch_with_closure(self, closure: Closure):
        workspace = self._context.get_workspace()
        file_to_patch = Path(
            os.path.join(
                workspace, "prod_00", closure.machine.name, "startup", "sub_start.sh"
            )
        )
        lines = file_to_patch.read_text(encoding="utf-8")
        patched_script = re.sub(
            r"\bpython3\b",
            closure.python_path,
            lines,
        )
        file_to_patch.write_text(patched_script, encoding="utf-8")

    def get_provision_path_for_machine(self, machine: Machine) -> str:
        return Path(os.path.join(self.deployed_files_path, machine.name)).as_posix()

    @property
    def deployed_files_path(self) -> str:
        return self._context.get_result_location()


@dataclass(frozen=True)
class NvidiaFlare:
    flare_project_file: Path
    refedez_folder: Path
    workspace_name: str = "workspace"

    def __post_init__(self):
        os.makedirs(self.workspace_folder, exist_ok=True)

    @property
    def workspace_folder(self) -> Path:
        return Path(os.path.join(self.refedez_folder, self.workspace_name))

    def provision(self) -> FlareProvision:
        workspace = self.workspace_folder.resolve().as_posix()
        project = self.flare_project_file.resolve().as_posix()
        # Same implementation as:
        # from nvflare.lighter.provision import provision
        # But we do it manually because we want to get the Context
        project_dict = load_yaml(project)
        project = prepare_project(project_dict, None, None)
        builders = prepare_builders(project_dict)
        provisioner = Provisioner(workspace, builders)
        context = provisioner.provision(project)
        return FlareProvision(_context=context)


def nvidia_flare_from_configuration(configuration: ReFedEzProject) -> NvidiaFlare:
    return NvidiaFlare(
        flare_project_file=configuration.config.flare_config,
        refedez_folder=configuration.config.refedez_folder,
    )



import dataclasses
import json
from dataclasses import dataclass
from pathlib import Path
from typing import List

from dacite import from_dict

from refedez.build.nix import Closure
from refedez.common.errors import InvalidStateError


@dataclass(frozen=True)
class MachineRunning:
    closure: Closure
    tmp_dir: str


class State:
    pass


@dataclass
class Clean(State):
    pass


@dataclass
class Dirty(State):
    pass


@dataclass
class Running(State):
    machines: List[MachineRunning]
    deployed_files_path: str


@dataclass
class StateFile:
    directory: Path
    STATE_MAP = {"clean": Clean, "running": Running, "dirty": Dirty}

    @property
    def state_file_path(self) -> str:
        return f"{self.directory.resolve().as_posix()}/state.json"

    @property
    def exists(self) -> bool:
        return Path(self.state_file_path).exists()

    def __get_key(self, state: State):
        for k, v in self.STATE_MAP.items():
            if isinstance(state, v):
                return k
        raise NotImplementedError(f"No key for state {state} found")

    def write_state(self, state: State):
        key = self.__get_key(state)
        result = {
            "type": key,
            "data": dataclasses.asdict(state),
        }
        with open(self.state_file_path, "w", encoding="utf-8") as f:
            json.dump(result, f, ensure_ascii=False, indent=4)

    @property
    def state(self) -> State:
        if not self.exists:
            return Clean()
        with open(self.state_file_path, "r", encoding="utf-8") as f:
            data = json.loads(f.read())
            type = data.get("type")
            state_data = data.get("data")
            return from_dict(data_class=self.STATE_MAP[type], data=state_data)


@dataclass(frozen=True)
class ProjectState:
    file: StateFile

    @property
    def is_running(self) -> bool:
        match self.state:
            case Running(_):
                return True
            case _:
                return False

    @property
    def state(self) -> State:
        return self.file.state

    def update_state(self, state: State):
        match self.state, state:
            case Running(_), Running(_):
                raise InvalidStateError()
        self.file.write_state(state)

    @staticmethod
    def from_path(state_directory: Path):
        return ProjectState(file=StateFile(directory=state_directory))



from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, Any, List

from refedez.lib.algorithms import FedAlgorithm


class Backend(Enum):
    PYTORCH = "torch"
    NUMPY = "numpy"


@dataclass(frozen=True)
class Client:
    """Configuration for a federated learning client.

    Represents a client participant in the federated learning system, including
    its name and any environment variables required for execution.

    Attributes:
        name: The unique name identifier for the client.
        env_vars: Dictionary of environment variables to set for the client.
    """

    name: str
    env_vars: Dict[str, str] = field(default_factory=dict)


@dataclass(frozen=True)
class Server:
    """Configuration for the federated learning server.

    Represents the central server in the federated learning system, responsible
    for coordinating the training process and aggregating model updates.

    Attributes:
        name: The unique name identifier for the server.
        save_model_path: Optional path where the trained model should be saved.
    """

    name: str
    save_model_path: str | None = None


@dataclass(frozen=True)
class JobConfiguration:
    server: Server
    clients: List[Client]
    refedez_config: str
    to_execute_script_path: str
    num_rounds: int
    model_cls: Any
    type: Backend
    algorithm: FedAlgorithm

    @property
    def num_clients(self) -> int:
        return len(self.clients)



import os
import socket
from dataclasses import dataclass
from typing import Any

from paramiko import (
    SSHClient,
    BadHostKeyException,
    AuthenticationException,
    SSHException,
    SFTPClient,
)
from paramiko.ssh_exception import UnableToAuthenticate, NoValidConnectionsError

from refedez.common.errors import SshConnectionError, SshCommandError
from refedez.common.machines import RemoteCredentials, Machine


class SFTPConnection:
    def __init__(self, sftp: SFTPClient):
        self._sftp = sftp

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        try:
            self._sftp.close()
        except Exception:
            pass

    def mkdir(self, remote_path: str, ignore_existing: bool = True):
        try:
            self._sftp.mkdir(remote_path)
        except Exception:
            if ignore_existing:
                # try to stat it to see if it exists
                try:
                    self._sftp.stat(remote_path)
                except Exception:
                    # it truly failed
                    raise
                # else it exists, so ignore
            else:
                raise

    def _ensure_remote_dir(self, remote_path: str):
        """
        Create all components in remote_path (mkdir -p style).
        """
        parts = remote_path.strip("/").split("/")
        cur = "/" if remote_path.startswith("/") else ""
        for part in parts:
            if not part:
                continue
            if cur == "/":
                cur = "/" + part
            elif cur == "":
                cur = part
            else:
                cur = cur.rstrip("/") + "/" + part
            self.mkdir(cur, ignore_existing=True)

    def put_folder(self, local_folder: str, remote_folder: str):
        local_folder = os.path.normpath(local_folder)
        self._ensure_remote_dir(remote_folder)

        for root, dirs, files in os.walk(local_folder):
            rel = os.path.relpath(root, local_folder)
            # remote path for this root
            if rel == "." or rel == "":
                remote_root = remote_folder
            else:
                # convert local separators to '/'
                rel_posix = rel.replace(os.sep, "/")
                remote_root = remote_folder.rstrip("/") + "/" + rel_posix

            # make subdirectories
            for d in dirs:
                remote_sub = remote_root.rstrip("/") + "/" + d
                self.mkdir(remote_sub, ignore_existing=True)

            # put files
            for f in files:
                local_f = os.path.join(root, f)
                remote_f = remote_root.rstrip("/") + "/" + f
                self._sftp.put(local_f, remote_f)


@dataclass
class SSHConnection:
    credentials: RemoteCredentials
    client: SSHClient | None = None

    def __enter__(self) -> "SSHConnection":
        try:
            self.client = SSHClient()
            self.client.load_system_host_keys()
            self.client.connect(
                hostname=self.credentials.ip, username=self.credentials.user
            )
            return self
        except (
            IOError,
            BadHostKeyException,
            AuthenticationException,
            UnableToAuthenticate,
            socket.error,
            NoValidConnectionsError,
            SSHException,
        ):
            raise SshConnectionError()

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.client:
            self.client.close()
            self.client = None

    @staticmethod
    def from_machine(machine: Machine) -> "SSHConnection":
        return SSHConnection(credentials=machine.credentials)

    def exec_command(self, command: str) -> Any:
        if self.client:
            stdin, stdout, stderr = self.client.exec_command(command)
            exit_status = stdout.channel.recv_exit_status()
            if exit_status == 0:
                return (
                    None,
                    "\n".join(stdout.readlines()),
                    "\n".join(stderr.readlines()),
                )
            else:
                raise SshCommandError(
                    f"Command failed with exit status {exit_status}, stdout: {stdout.readlines()}, stderr: {stderr.readlines()}"
                )
        raise SshCommandError("Client is not enabled...")

    def get_sftp(self) -> SFTPConnection:
        if self.client is None:
            raise ValueError("TODO: Change this")
        sftp = self.client.open_sftp()
        return SFTPConnection(sftp)



import os
from abc import ABC
from typing import Any

from nvflare import FedJob
from nvflare.app_common.np.np_model_persistor import NPModelPersistor
from nvflare.app_common.workflows.fedavg import FedAvg
from nvflare.client.config import ExchangeFormat
from nvflare.job_config.script_runner import ScriptRunner, FrameworkType

from refedez.lib.algorithms.common import from_algorithm
from refedez.lib.config import JobConfiguration
from refedez.lib.constants import ENV_RUN_JOB_STAGE


class FederatedNumpy(ABC):
    """Abstract base class for federated learning models using NumPy.

    This class defines the interface for models that participate in federated learning
    workflows using NumPy arrays for weight representation.

    Subclasses must implement the abstract methods to define the model's forward pass,
    weight retrieval and setting, and training step logic.
    """

    def forward(self, x: Any):
        raise NotImplementedError()

    def get_weights(self) -> Any:
        raise NotImplementedError()

    def set_weights(self, new_weights: Any) -> None:
        raise NotImplementedError()

    def train_step(self, learning_rate=1.0) -> Any:
        raise NotImplementedError()


def __create_numpy_job(config: JobConfiguration, output_path: str):
    name = "fed_numpy"
    job = FedJob(name=name, min_clients=config.num_clients)

    # TODO: Change
    persistor = NPModelPersistor(model_dir="/models", model_name="server.npy")
    persistor_id = job.to_server(persistor, config.server.name)

    controller = from_algorithm(config.algorithm, config, persistor_id).get_controller()
    job.to_server(controller, config.server.name)

    # Define script runner for clients
    train_script = config.to_execute_script_path
    # TODO: Add environment flag for running
    script_runner = ScriptRunner(
        script=train_script,
        script_args=f"{ENV_RUN_JOB_STAGE}",
        launch_external_process=False,
        framework=FrameworkType.NUMPY,
        server_expected_format=ExchangeFormat.NUMPY,
        params_transfer_type="FULL",
    )

    # Assign to clients
    for client in config.clients:
        client_name = client.name
        job.to(script_runner, client_name, tasks=["train"])

    # Export job folder
    job.export_job(job_root=output_path)
    return os.path.join(output_path, name)


def __execute(model: FederatedNumpy):
    pass



from typing import List

from rich import print
from rich.table import Table

from refedez.common.config import get_refedez_config_from_yaml
from refedez.common.project import ReFedEzProject
from refedez.common.state import Running, Clean, MachineRunning, State, Dirty


def print_running_machines(machines: List[MachineRunning], deployed_files_path: str):
    print(f"[green]Based on workspace: [/green] {deployed_files_path}")
    table = Table("Machine Name", "Type", "Files Folder", "IP", "User")
    for machine in machines:
        m = machine.closure.machine
        type = m.connection.type
        ip = m.connection.remote_credentials.ip if m.is_remote else "N/A"
        user = m.connection.remote_credentials.user if m.is_remote else "N/A"
        table.add_row(m.name, type, machine.tmp_dir, ip, user)
    print(table)


def print_from_state(state: State):
    match state:
        case Running(machines, deployed_files_path):
            print_running_machines(machines, deployed_files_path)
        case Clean():
            print("[green]Project is clean![/green]")
        case Dirty():
            print("[orange]Project contains old builds and unnecessary files[/orange]")
        case _:
            print(f"[red]Invalid state... {state}[/red]")


def status(path: str) -> None:
    configuration = get_refedez_config_from_yaml(path=path)
    project = ReFedEzProject(config=configuration)
    project_state = project.project_state
    print_from_state(state=project_state.state)



from dataclasses import dataclass
from pathlib import Path
from typing import Generator, Tuple

from refedez.common.config import Config
from refedez.common.machines import Machine
from refedez.build.uv import UvProject
from refedez.common.state import ProjectState


@dataclass(frozen=True)
class ReFedEzProject:
    config: Config

    def __post_init__(self):
        # TODO: Check that we have the same machines as nvflare on our configuration
        pass

    @property
    def machines_with_project(
        self,
    ) -> Generator[Tuple[Machine, UvProject | None], None, None]:
        for machine in self.config.machines:
            machine = machine.to_machine()
            yield machine, self.__get_uv_project_for(machine)

    @property
    def num_machines(self) -> int:
        return len(self.config.machines)

    def __get_uv_project_for(self, machine: Machine) -> UvProject | None:
        for project in self.config.project:
            if project.name == machine.name:
                return UvProject(path=Path(project.folder))
        return None

    @property
    def project_folder(self) -> Path:
        return self.config.refedez_folder

    @property
    def project_state(self) -> ProjectState:
        return ProjectState.from_path(state_directory=self.project_folder)



import sys
from inspect import getfile
from typing import List

from refedez.lib.algorithms import FedAlgorithm
from refedez.lib.backends import Backend
from refedez.lib.backends.numpy import __execute as execute_numpy
from refedez.lib.backends.torch import __execute as execute_torch
from refedez.lib.config import JobConfiguration, Server, Client
from refedez.lib.constants import ENV_RUN_JOB_STAGE
from refedez.lib.job import execute_job_config


def Federated(
        server: Server,
        clients: List[Client],
        refedez_config: str,
        num_rounds=1,
        backend=Backend.PYTORCH,
        algorithm=FedAlgorithm.FED_AVG,
):
    """
     Runs a federated learning experiment using the specified backend and configuration.

    This function orchestrates the communication and training rounds between a central
    server and multiple clients according to the provided `refedez_config` file.
    The backend determines the machine learning framework used (e.g., PyTorch or NumPy).

    Args:
        server (Server): The federated server responsible for aggregating model updates
            and coordinating the training process.
        clients (List[Client]): A list of participating clients, each responsible for
            local model training and update submission.
        refedez_config (str): Path to the ReFedEz configuration file (YAML or JSON)
            that defines the experiment setup, data paths, and hyperparameters.
        num_rounds (int, optional): Number of federated training rounds to execute.
            Defaults to 1.
        backend (Backend, optional): Backend to use for computation. Can be one of the
            supported frameworks in `Backend` (e.g., `Backend.PYTORCH`, `Backend.NUMPY`).
            Defaults to `Backend.PYTORCH`.
        algorithm (FedAlgorithm, optional): Backend to use for computation. Can be one of the
            supported frameworks in `FedAlgorithm` (e.g., `FedAlgorithm.FED_AVG`, `FedAlgorithm.SCAFFOLD`).
            Defaults to `FedAlgorithm.FED_AVG`.

    Returns:
        None: This function does not return a value directly, but performs training,
        logging, and potentially saves model artifacts or metrics as side effects.

    Raises:
        ValueError: If the configuration file path is invalid or unreadable.
        RuntimeError: If the backend fails to initialize or training fails mid-process.
        ConnectionError: If communication with clients fails during aggregation rounds.

    Example:
    """

    def decorator(cls):
        # Only run the logic if this module is executed as __main__
        if cls.__module__ == "__main__":
            args = sys.argv
            should_execute = len(args) >= 2 and ENV_RUN_JOB_STAGE in args[1]

            if should_execute:
                if backend == Backend.PYTORCH:
                    execute_torch(cls())
                elif backend == Backend.NUMPY:
                    execute_numpy(cls())
                else:
                    raise NotImplementedError(f"Unsupported backend {backend}")
            else:
                script_path = getfile(cls)
                job_config = JobConfiguration(
                    server=server,
                    clients=clients,
                    refedez_config=refedez_config,
                    to_execute_script_path=script_path,
                    num_rounds=num_rounds,
                    model_cls=cls,
                    type=backend,
                    algorithm=algorithm
                )
                execute_job_config(job_config)
                exit(0)
        return cls

    return decorator



import uuid
from dataclasses import dataclass
from pathlib import Path


@dataclass(frozen=True)
class UvProject:
    path: Path

    def __post_init__(self):
        # TODO: Here in init assert that we have a project.yml and a uv.lock file inside the Path
        pass

    @property
    def hash(self) -> str:
        # TODO: Put a hash here
        return str(uuid.uuid4())



# Config Related


class NvidiaFlareProjectFileNotFound(Exception):
    pass


class MachineNotFound(Exception):
    pass


class NotAValidConnectionType(Exception):
    pass


# Uv related


class MachineDoesNotHaveAValidUvProject(Exception):
    pass


# Nix Related (Install it, build closure, etc)


class NixBuildClosureError(Exception):
    message: str

    def __init__(self, message: str):
        super().__init__()
        self.message = message


class NixClosureAlreadyUsed(Exception):
    message: str

    def __init__(self, message: str):
        super().__init__()
        self.message = message


class NixInstallationError(Exception):
    message: str


# Ssh Related


class SshConnectionError(Exception):
    pass


class SshCommandError(Exception):
    pass


# State


class ProjectAlreadyRunningError(Exception):
    pass


class InvalidStateError(Exception):
    pass



import os
import subprocess
from dataclasses import dataclass
from shutil import which

from refedez.build.ssh import SSHConnection
from refedez.build.uv import UvProject
from refedez.common.errors import NixBuildClosureError, SshCommandError
from refedez.common.machines import Machine

NIX_BUILD_UV_PROJECT_TEMPLATE = "github:pablito2020/testing-flakes-template"


@dataclass
class Closure:
    identifier: str
    machine: Machine
    src_uv_project_path: str | None = None

    def __post_init__(self):
        # TODO: Check if identifier is really from nix store
        self.identifier = self.identifier.strip()

    @property
    def python_path(self) -> str:
        return f"{self.identifier}/bin/python3"


def build_pyproject_from_project(project: UvProject | None) -> str:
    command = (
        f"nix build {NIX_BUILD_UV_PROJECT_TEMPLATE} --no-link --print-out-paths"
        if project is None
        else f"nix build {NIX_BUILD_UV_PROJECT_TEMPLATE} --override-input mysrc path:{project.path.resolve()} --no-link --print-out-paths"
    )
    args = command.split()
    try:
        output_path_closure = subprocess.run(args, capture_output=True, check=True)
        return output_path_closure.stdout.decode("utf-8")
    except subprocess.CalledProcessError as e:
        print(e.stdout)
        print(e.stderr)
        raise NixBuildClosureError(
            message=f"Return Code: {e.returncode}. \n Stdout Message: {e.stdout}. \n Stderr Message: {e.stderr}"
        )


@dataclass(frozen=True)
class NixMachine:
    machine: Machine

    def has_nix_installed(self) -> bool:
        if self.machine.is_local:
            return which("nix") is not None
        with SSHConnection.from_machine(self.machine) as conn:
            try:
                conn.exec_command("nix --version")
                return True
            except SshCommandError:
                return False

    def install_nix(self):
        if self.has_nix_installed():
            return
        install_script = "curl -fsSL https://install.determinate.systems/nix | sh -s -- install --no-confirm"
        match self.machine.is_local:
            case True:
                os.system(install_script)
            case False:
                # TODO: See if we can catch errors?
                with SSHConnection.from_machine(self.machine) as conn:
                    stdin, stdout, stderr = conn.exec_command(install_script)

    def build_closure(self, project: UvProject | None) -> Closure:
        if self.machine.is_local:
            closure_result = build_pyproject_from_project(project)
            return Closure(
                identifier=closure_result,
                machine=self.machine,
                src_uv_project_path=project.path.as_posix() if project else None,
            )
        with SSHConnection.from_machine(self.machine) as conn:
            command = (
                f"nix build {NIX_BUILD_UV_PROJECT_TEMPLATE} --no-link --print-out-paths"
            )
            if project:
                with conn.get_sftp() as sftp:
                    directory_client = f"/tmp/{project.hash}"
                    sftp.put_folder(
                        local_folder=project.path.as_posix(),
                        remote_folder=directory_client,
                    )
                command = f"nix build {NIX_BUILD_UV_PROJECT_TEMPLATE} --override-input mysrc path:{directory_client} --no-link --print-out-paths"
            # TODO: Check for error
            stdin, stdout, stderr = conn.exec_command(command)
            return Closure(
                identifier=stdout,
                src_uv_project_path=directory_client,
                machine=self.machine,
            )



from refedez.lib.config import Server, Client
from refedez.lib import Federated
from refedez.lib.backends.numpy import FederatedNumpy
from refedez.lib.backends.torch import FederatedTorch

__all__ = ["Federated", "FederatedNumpy", "FederatedTorch", Server, Client]



from refedez.lib.backends.numpy import FederatedNumpy
from refedez.lib.backends.torch import FederatedTorch
from refedez.lib.config import Backend


def get_backend_from_type(cls):
    if issubclass(cls, FederatedTorch):
        return Backend.PYTORCH
    elif issubclass(cls, FederatedNumpy):
        return Backend.NUMPY
    else:
        raise NotImplementedError(f"ReFedEz doesn't support {cls} types")



from refedez.lib.refedez import Federated
from refedez.lib.backends.numpy import FederatedNumpy
from refedez.lib.backends.torch import FederatedTorch
from refedez.lib.config import Server, Client

__all__ = ["Federated", "FederatedNumpy", "FederatedTorch", "Server", "Client"]



import dataclasses
import os
import shutil
from dataclasses import dataclass
from tempfile import mkdtemp

from nvflare.fuel.flare_api.flare_api import new_secure_session
from nvflare.tool.job.job_cli import (
    get_app_dirs_from_job_folder,
    prepare_job_config,
    find_admin_user_and_dir,
)
from nvflare.tool.job.job_client_const import DEFAULT_APP_NAME

from refedez.common.config import get_refedez_config_from_yaml
from refedez.common.project import ReFedEzProject
from refedez.common.state import Running
from refedez.lib.backends import Backend
from refedez.lib.backends.numpy import __create_numpy_job
from refedez.lib.backends.torch import __create_torch_job
from refedez.lib.config import JobConfiguration


@dataclass(frozen=True)
class Job:
    id: str


def internal_submit_job(admin_user_dir, username, temp_job_dir) -> str:
    # From: https://github.com/NVIDIA/NVFlare/blob/2.6/nvflare/tool/job/job_cli.py#L347
    print("trying to connect to the server")
    sess = new_secure_session(username=username, startup_kit_location=admin_user_dir)
    job_id = sess.submit_job(temp_job_dir)
    print(f"job: '{job_id} was submitted")
    return job_id


@dataclass
class Args:
    job_folder: str
    config_file: str | None = None

    def __iter__(self):
        return iter(dataclasses.asdict(self))


def submit_job(job_folder: str) -> Job:
    # From: https://github.com/NVIDIA/NVFlare/blob/2.6/nvflare/tool/job/job_cli.py#L347
    cmd_args = Args(job_folder=job_folder)
    temp_job_dir = None
    try:
        if not os.path.isdir(cmd_args.job_folder):
            raise ValueError(f"invalid job folder: {cmd_args.job_folder}")

        temp_job_dir = mkdtemp()
        shutil.copytree(cmd_args.job_folder, temp_job_dir, dirs_exist_ok=True)

        app_dirs = get_app_dirs_from_job_folder(cmd_args.job_folder)
        app_names = [os.path.basename(f) for f in app_dirs]
        app_names = app_names if app_names else [DEFAULT_APP_NAME]

        prepare_job_config(cmd_args, app_names, temp_job_dir)
        admin_username, admin_user_dir = find_admin_user_and_dir()

        # TODO: get from nvflare project
        admin_username = "admin@pablofraile.net"

        return Job(internal_submit_job(admin_user_dir, admin_username, temp_job_dir))
    except ValueError as e:
        print(f"\nUnable to submit due to: {e} \n")
    finally:
        if temp_job_dir:
            shutil.rmtree(temp_job_dir)


def create_job(job_config: JobConfiguration, output_path: str) -> str:
    match job_config.type:
        case Backend.PYTORCH:
            return __create_torch_job(job_config, output_path)
        case Backend.NUMPY:
            return __create_numpy_job(job_config, output_path)


def execute_job_config(job_config: JobConfiguration):
    configuration = get_refedez_config_from_yaml(path=job_config.refedez_config)
    project = ReFedEzProject(config=configuration)
    state = project.project_state
    job_folder = f"{project.project_folder}/jobs"
    match state.state:
        case Running(_, deployed_files_path):
            # TODO: Check that project and job_config are valid (names, etc)
            # This is needed for flare to detect our current deployment
            os.environ["NVFLARE_STARTUP_KIT_DIR"] = deployed_files_path
            created_folder = create_job(job_config, job_folder)
            job = submit_job(created_folder)
            print(f"Started job: {job}")
        case _:
            # TODO: Change for another exception
            raise ValueError("Project should be running..")



import os
from typing import Any

import nvflare.client as flare
import torch
from nvflare import FedJob
from nvflare.app_common.workflows.fedavg import FedAvg
from nvflare.app_opt.pt import PTFileModelPersistor
from nvflare.client.config import ExchangeFormat
from nvflare.job_config.script_runner import ScriptRunner, FrameworkType
from torch import nn

from refedez.lib.algorithms.common import from_algorithm
from refedez.lib.config import JobConfiguration
from refedez.lib.constants import ENV_RUN_JOB_STAGE


class FederatedTorch(nn.Module):
    """Base class for federated learning models using PyTorch.

    This class extends PyTorch's nn.Module and provides the interface for models
    in federated learning scenarios. It handles the initialization and defines
    abstract methods for weight management and training.

    Subclasses should implement the forward pass and training logic specific to
    their model architecture.
    """

    def __init__(self):
        super(FederatedTorch, self).__init__()

    def forward(self, x: Any):
        raise NotImplementedError()

    def get_weights(self) -> torch.Tensor:
        raise NotImplementedError()

    def set_weights(self, new_weights: torch.Tensor) -> None:
        raise NotImplementedError()

    def train_step(self, learning_rate=1.0) -> torch.Tensor:
        raise NotImplementedError()


def __create_torch_job(config: JobConfiguration, output_path: str):
    # TODO: Maybe add timestamp to job?
    name = "torch_job"
    job = FedJob(name=name, min_clients=config.num_clients)

    persistor = PTFileModelPersistor(
        model=config.model_cls(), global_model_file_name=config.server.save_model_path
    )
    persistor_id = job.to_server(persistor, config.server.name)

    # Controller workflow
    controller = from_algorithm(config.algorithm, config, persistor_id).get_controller()
    job.to_server(controller, config.server.name)  # replace with your server name

    train_script = config.to_execute_script_path
    for client in config.clients:
        # TODO: Build env vars + add the env var for ignoring creation of job
        script_runner = ScriptRunner(
            script=train_script,
            script_args=f"{ENV_RUN_JOB_STAGE}",
            launch_external_process=False,
            framework=FrameworkType.PYTORCH,
            server_expected_format=ExchangeFormat.PYTORCH,
            params_transfer_type="FULL",
        )
        client_name = client.name
        job.to(script_runner, client_name, tasks=["train"])

    job.export_job(job_root=output_path)
    return os.path.join(output_path, name)


def __execute(model: FederatedTorch):
    # Initialize FLARE
    flare.init()
    sys_info = flare.system_info()
    client_name = sys_info["site_name"]
    print(f"Client {client_name} initialized")

    while flare.is_running():
        # Receive model from server
        input_model = flare.receive()
        print(f"Client {client_name}, current_round={input_model.current_round}")
        print(f"Received weights: {input_model.params}")

        # Load the received model weights
        if input_model.params == {}:
            params = model.get_weights()
        else:
            params = input_model.params
        model.set_weights(params)

        # Perform local training
        print(f"Client {client_name} starting training...")
        new_params = model.train_step(learning_rate=1.0)

        print(
            f"Client {client_name} finished training for round {input_model.current_round}"
        )
        print(f"Sending weights: {new_params}")

        output_model = flare.FLModel(
            params=model.cpu().state_dict(),
            params_type="FULL",
            current_round=input_model.current_round,
        )
        flare.send(output_model)



import typer

from refedez.cli.commands.start import start as start_impl
from refedez.cli.commands.status import status as status_impl
from refedez.cli.commands.stop import stop as stop_impl
from refedez.cli.commands.clean import clean as clean_impl

app = typer.Typer()

DEFAULT_CONFIG_FILE = "./refedez.yaml"
DEFAULT_PATH_FLAG = typer.Option(DEFAULT_CONFIG_FILE, help="Path to the config file")


@app.command()
def start(path: str = DEFAULT_PATH_FLAG):
    start_impl(path=path)


@app.command()
def status(path: str = DEFAULT_PATH_FLAG):
    status_impl(path=path)


@app.command()
def stop(path: str = DEFAULT_PATH_FLAG):
    stop_impl(path=path)


@app.command()
def clean(path: str = DEFAULT_PATH_FLAG):
    clean_impl(path=path)



import shutil
import subprocess
import uuid
from typing import List

from tqdm import tqdm

from refedez.build.flare import FlareProvision, nvidia_flare_from_configuration
from refedez.build.nix import NixMachine, Closure
from refedez.build.ssh import SSHConnection
from refedez.build.uv import UvProject
from refedez.common.config import get_refedez_config_from_yaml
from refedez.common.errors import ProjectAlreadyRunningError
from refedez.common.machines import Machine
from refedez.common.project import ReFedEzProject
from refedez.common.state import MachineRunning, Running


def build_closure_for_machine(
    machine: Machine, uv_project: UvProject | None
) -> Closure:
    nix_machine = NixMachine(machine)
    nix_machine.install_nix()
    return nix_machine.build_closure(uv_project)


def build_closures_for_all_machines(project: ReFedEzProject) -> List[Closure]:
    builds = []
    with tqdm(
        total=project.num_machines,
        desc="📦 Building environments...",
        bar_format="{l_bar}{bar} | {n_fmt}/{total_fmt} [{elapsed}<{remaining}]",
    ) as pbar:
        for machine, project in project.machines_with_project:
            builds.append(
                build_closure_for_machine(machine=machine, uv_project=project)
            )
            pbar.set_postfix_str(f"🖥️ {machine.name}")
            pbar.update(1)
    return builds


def patch_provision_with_closures(provision: FlareProvision, closures: List[Closure]):
    with tqdm(
        total=len(closures),
        desc="✂️ Patching provisioned files so they run on the generated ReFedEz environment...",
        bar_format="{l_bar}{bar} | {n_fmt}/{total_fmt} [{elapsed}<{remaining}]",
    ) as pbar:
        for closure in closures:
            provision.patch_with_closure(closure)
            pbar.update(1)


def upload_and_run_build(
    provision: FlareProvision, closures: List[Closure]
) -> List[MachineRunning]:
    # TODO: If something fails, then maybe kill the others?
    running_machines = []
    with tqdm(
        total=len(closures),
        desc="🚀 Starting services on all machines...",
        bar_format="{l_bar}{bar} | {n_fmt}/{total_fmt} [{elapsed}<{remaining}]",
    ) as pbar:
        for closure in closures:
            path = provision.get_provision_path_for_machine(closure.machine)
            identifier = str(uuid.uuid4())
            directory_client = f"/tmp/{identifier}"
            command = f"nohup {directory_client}/startup/start.sh > /dev/null 2>&1 &"
            match closure.machine.is_local:
                case True:
                    shutil.copytree(path, directory_client)
                    subprocess.run(command.split(), check=True)
                case False:
                    with SSHConnection.from_machine(closure.machine) as conn:
                        with conn.get_sftp() as sftp:
                            sftp.put_folder(
                                local_folder=path, remote_folder=directory_client
                            )
                        conn.exec_command(f"chmod -R +x {directory_client}")
                        conn.exec_command(command)
            running_machines.append(
                MachineRunning(closure=closure, tmp_dir=directory_client)
            )
            pbar.update(1)
    return running_machines


def execute(refedez_project: ReFedEzProject):
    state = refedez_project.project_state
    if state.is_running:
        raise ProjectAlreadyRunningError()
    flare = nvidia_flare_from_configuration(refedez_project)
    provision = flare.provision()
    closures = build_closures_for_all_machines(project=refedez_project)
    patch_provision_with_closures(provision, closures)
    machines_running = upload_and_run_build(provision, closures)
    state.update_state(
        Running(
            machines=machines_running, deployed_files_path=provision.deployed_files_path
        )
    )


def start(path: str) -> None:
    configuration = get_refedez_config_from_yaml(path=path)
    project = ReFedEzProject(config=configuration)
    execute(project)



import os
from pathlib import Path
from typing import List

from refedez.build.ssh import SSHConnection
from refedez.common.config import get_refedez_config_from_yaml
from refedez.common.errors import InvalidStateError
from refedez.common.machines import Machine
from refedez.common.project import ReFedEzProject
from refedez.common.state import Running, MachineRunning, Dirty

# The script provisioned by nvflare provision uses this for detecting it has
# to stop the proces... So yeah, it is what it is
SHUTDOWN_FILE = "shutdown.fl"


def stop_machine_remote(machine: Machine, directory: str):
    command = f"touch {directory}/{SHUTDOWN_FILE}"
    with SSHConnection.from_machine(machine) as conn:
        conn.exec_command(command)


def stop_machine_local(directory: str):
    file = Path(os.path.join(directory, SHUTDOWN_FILE))
    if not file.exists():
        file.touch()


def stop_machines(machines: List[MachineRunning]):
    for machine in machines:
        m = machine.closure.machine
        if m.is_remote:
            stop_machine_remote(m, machine.tmp_dir)
        else:
            stop_machine_local(machine.tmp_dir)


def stop(path: str):
    configuration = get_refedez_config_from_yaml(path=path)
    refedez_project = ReFedEzProject(config=configuration)
    refedez_state = refedez_project.project_state
    match refedez_state.state:
        case Running(machines, _):
            stop_machines(machines)
            refedez_state.update_state(Dirty())
        case _:
            raise InvalidStateError()



import os
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional
import yaml

from refedez.common.errors import NvidiaFlareProjectFileNotFound
from refedez.common.machines import Machine, Connection, RemoteCredentials


@dataclass(frozen=True)
class ReFedEzConfig:
    folder: str
    capabilities: str


@dataclass(frozen=True)
class MachineConfig:
    name: str
    type: str
    ip: Optional[str] = None
    user: Optional[str] = None

    def to_machine(self) -> Machine:
        remote_credentials = None
        if self.ip or self.user:
            remote_credentials = RemoteCredentials(user=self.user, ip=self.ip)
        return Machine(
            name=self.name,
            connection=Connection(
                type=self.type, remote_credentials=remote_credentials
            ),
        )


@dataclass(frozen=True)
class ProjectConfig:
    name: str
    folder: str


@dataclass(frozen=True)
class Config:
    refedez: ReFedEzConfig
    machines: List[MachineConfig]
    project: List[ProjectConfig]

    def __post_init__(self):
        os.makedirs(self.refedez.folder, exist_ok=True)
        if not os.path.exists(self.refedez.capabilities):
            raise NvidiaFlareProjectFileNotFound()

    @property
    def refedez_folder(self) -> Path:
        return Path(self.refedez.folder)

    @property
    def flare_config(self) -> Path:
        return Path(self.refedez.capabilities)


def get_refedez_config_from_yaml(path: str) -> Config | None:
    with open(path, "r") as f:
        data = yaml.safe_load(f)
    refedez = ReFedEzConfig(**data["refedez"])
    machines = [
        MachineConfig(name=name, **attrs)
        for name, attrs in data.get("machines", {}).items()
    ]
    project = [
        ProjectConfig(name=name, **attrs)
        for name, attrs in data.get("project", {}).items()
    ]
    return Config(refedez=refedez, machines=machines, project=project)



from dataclasses import dataclass

from nvflare.app_common.workflows.model_controller import ModelController
from nvflare.app_common.workflows.scaffold import Scaffold

from refedez.lib.algorithms.common import FedAlgorithmBase
from refedez.lib.config import JobConfiguration


@dataclass(frozen=True)
class ScaffoldAlgorithm(FedAlgorithmBase):
    num_clients: int
    num_rounds: int
    persistor_id: str

    def get_controller(self)-> ModelController:
        return Scaffold(
            num_clients=self.num_clients,
            num_rounds=self.num_rounds,
            persistor_id=self.persistor_id,
        )

    @staticmethod
    def from_(config: JobConfiguration, persistor_id: str) -> 'ScaffoldAlgorithm':
        return ScaffoldAlgorithm(
            num_clients=config.num_clients,
            num_rounds=config.num_rounds,
            persistor_id=persistor_id
        )




import shutil

from refedez.common.config import get_refedez_config_from_yaml
from refedez.common.errors import InvalidStateError
from refedez.common.project import ReFedEzProject
from refedez.common.state import Dirty


def clean(path: str):
    configuration = get_refedez_config_from_yaml(path=path)
    refedez_project = ReFedEzProject(config=configuration)
    refedez_state = refedez_project.project_state
    match refedez_state.state:
        case Dirty():
            shutil.rmtree(refedez_project.project_folder)
            pass
        case _:
            raise InvalidStateError()



from dataclasses import dataclass

import torch
from nvflare.app_common.workflows.fedavg import FedAvg
from nvflare.app_common.workflows.model_controller import ModelController
from nvflare.app_opt.pt.fedopt_ctl import FedOpt

from refedez.lib.algorithms.common import FedAlgorithmBase
from refedez.lib.config import JobConfiguration, Backend


@dataclass(frozen=True)
class FedOptAlgorithm(FedAlgorithmBase):
    source_model: torch.nn.Module

    def get_controller(self)-> ModelController:
        return FedOpt(source_model=self.source_model)

    @staticmethod
    def from_(config: JobConfiguration, persistor_id: str) -> 'FedOptAlgorithm':
        if config.type != Backend.PYTORCH:
            raise ValueError("FedOpt is only valid if you're using pytorch")
        return FedOptAlgorithm(
            source_model=config.model_cls
        )



from abc import ABC

from nvflare.app_common.workflows.model_controller import ModelController

from refedez.lib.algorithms import FedAlgorithm
from refedez.lib.algorithms.fed_avg import FedAvgAlgorithm
from refedez.lib.algorithms.fed_opt import FedOptAlgorithm
from refedez.lib.algorithms.scaffold import ScaffoldAlgorithm
from refedez.lib.config import JobConfiguration


class FedAlgorithmBase(ABC):

    def get_controller(self) -> ModelController:
        pass


def from_algorithm(algorithm: FedAlgorithm, config: JobConfiguration, persistor_id: str, ) -> FedAlgorithmBase:
    if FedAlgorithm.FED_AVG == algorithm:
        return FedAvgAlgorithm.from_(config, persistor_id)
    elif FedAlgorithm.FED_OPT == algorithm:
        return FedOptAlgorithm.from_(config, persistor_id)
    elif FedAlgorithm.SCAFFOLD == algorithm:
        return ScaffoldAlgorithm.from_(config, persistor_id)
    else:
        raise ValueError(f"Not supported algorithm: {algorithm}")



from dataclasses import dataclass
from decimal import InvalidOperation
from typing import Literal

from refedez.common.errors import NotAValidConnectionType


@dataclass(frozen=True)
class RemoteCredentials:
    user: str
    ip: str
    password: str | None = None
    port: int = 22


@dataclass(frozen=True)
class Connection:
    type: Literal["remote", "local"]
    remote_credentials: RemoteCredentials | None = None

    def __post_init__(self):
        if self.type not in ["remote", "local"]:
            raise NotAValidConnectionType()


@dataclass(frozen=True)
class Machine:
    name: str
    connection: Connection

    @property
    def is_remote(self):
        return self.connection.type == "remote"

    @property
    def is_local(self):
        return self.connection.type == "local"

    @property
    def credentials(self) -> RemoteCredentials:
        if self.is_local:
            raise InvalidOperation("A local machine can't have credentials.")
        return self.connection.remote_credentials
